const fs = require('fs'),
	path = require('path'),
	css = require('css'),
	mkdirp = require('mkdirp'),

	minifyCSS = require('./zc-minify-css.js'),
	minifyHTML = require('./zc-minify-html.js'),
	minifyXML = require('./zc-minify-xml.js'),
	minifyJS = require('./zc-minify-js.js'),
	Selectors = require('./zc-selectors.js');

module.exports = function(settings) {
	const MINIFY_SELECTORS = !!settings.minifySelectors,
		  MINIFY_JS = settings.minifyJS !== false,
		  MINIFY_HTML_JS = MINIFY_JS && !!settings.minifyHtmlJS,
		  MINIFY_HTML_CSS = !!settings.minifyHtmlCSS,
		  DEBUG = !!settings.debug,

		  FILE_EXTENSIONS = Object.assign({
			  js: ['js'],
			  css: ['css'],
			  html: ['html'],
			  xml: ['xml'],
		  }, settings.extensions);

	console.log("Debug: " + DEBUG);
	console.log("Minify selectors: " + MINIFY_SELECTORS);
	console.log("Minify JS: " + MINIFY_JS);
	console.log("Minify HTML JS: " + MINIFY_HTML_JS);
	console.log("Minify HTML CSS: " + MINIFY_HTML_CSS);

	if (MINIFY_SELECTORS) {
		var selectors = new Selectors({
				prepend: settings.selPrepend,
				externalSelectors: (() => {
					let ret = { c: [], i: [] };
					(settings.externalSelectors || []).forEach(selPkgName => {
						let file = `./${selPkgName}.json`,
							json = fs.readFileSync(file, { encoding: 'utf8' }),
							pkg = JSON.parse(json);

						[].push.apply(ret.c, pkg.c);
						[].push.apply(ret.i, pkg.i);

						console.log('Loaded selectors package %s', file);
					});
				})()
			});
	}

	// Clean paths and add slash at the end
	// NOTE: Undefined behaviour if src path doesn't exist; dst path doesn't have to exist
	const SRC_DIR = path.normalize(path.resolve(settings.src) + '/'),
		  DST_DIR = path.normalize(settings.dst + '/');

	console.log('Source: %s', SRC_DIR);
	console.log('Destination: %s', DST_DIR);

	// For files that only need to be copied and not altered, just copy them
	(settings.copy || []).forEach(fileToCopy => {
		let folder = path.normalize(fileToCopy.split('/').slice(0, -1).join('/') + '/');
		mkdirp.sync(DST_DIR + folder);
		fs.writeFileSync(DST_DIR + fileToCopy, fs.readFileSync(SRC_DIR + fileToCopy));
		console.log(`Copied ${fileToCopy}`);
	});

	var codeToMinify = [];

	// First, get and replace all the selectors from all files
	settings.files.forEach(file => {
		let _ = file.lastIndexOf('.') + 1 || file.length;

		/*
			WARNING: File paths are relative to src and build dirs
		*/
		let pathWithoutExt = file.slice(0, _), // "html/app." (last dot included) instead of "html/app.js"
			extension = file.slice(_), // js, html, css, etc.

			data = fs.readFileSync(SRC_DIR + file, { encoding: 'utf8' }); // The actual contents of the file

		data = data.replace(/<ZC-IMPORT\[(.+?)\]>/g, (_, importFile) => {
			var importFilePath;
			if (importFile.indexOf('/') === 0) {
				importFilePath = importFile;
			} else if (importFile.indexOf('/') > 0) {
				importFilePath = SRC_DIR + importFile;
			} else {
				if (extension) {
					importFilePath = SRC_DIR + pathWithoutExt + importFile + '.' + extension;
				} else {
					importFilePath = SRC_DIR + pathWithoutExt + '.' + importFile;
				}
			}
			return fs.readFileSync(importFilePath, { encoding: 'utf8' });
		});

		let zcVariables = {}, zcVariablesUnused = new Set();
		data = data.replace(/<ZC-SET\[(.+?)\]\[(.+?)\]>/g, (_, name, value) => {
			if (zcVariables[name] !== undefined) {
				throw new SyntaxError(`The zc variable ${name} has already been defined`);
			}
			zcVariables[name] = value;
			zcVariablesUnused.add(name);
			return '';
		});
		data = data.replace(/<ZC-USE\[(.+?)\]>/g, (_, name) => {
			let val = zcVariables[name];
			if (val === undefined) {
				throw new ReferenceError(`The zc variable ${name} is not defined`);
			}
			zcVariablesUnused.delete(name);
			return val;
		});
		for (let unusedVar of zcVariablesUnused.values()) {
			throw new SyntaxError(`The zc variable ${unusedVar} is not used`);
		}

		if (settings.onloadfile) data = settings.onloadfile(data, extension, file) || data;

		if (MINIFY_SELECTORS) {
			if (~FILE_EXTENSIONS.js.indexOf(extension)) {
				data = data.replace(/['"]([#.])([a-z-]+)['"]/g, (_, prefix, sel) => {
					let minifiedSel = selectors.minifySelector(sel, prefix == '.');
					return `"${prefix}${minifiedSel}"`;
				});
			} else if (~FILE_EXTENSIONS.css.indexOf(extension)) {
				let cssAst = css.parse(data),
					rulesToProcess = cssAst.stylesheet.rules,
					rule;

				while (rule = rulesToProcess.shift()) {
					if (rule.type == 'media') {
						rulesToProcess = rulesToProcess.concat(rule.rules);
					}

					if (rule.type == 'rule') {
						rule.selectors.forEach((ruleSelector, idx) => {
							rule.selectors[idx] = ruleSelector.replace(/([.#])([a-z-]+)(?=[ \[:,.#)]|$)/g, (_, prefix, sel) => {
								let minifiedSel = selectors.minifySelector(sel, prefix == '.');
								return `${prefix}${minifiedSel}`;
							});
						});
					}
				}
				data = css.stringify(cssAst, {
					compress: true,
				});
			} else if (~FILE_EXTENSIONS.html.indexOf(extension)) {
				data = data.replace(/ (class|for|id)="([a-z- ]+)"/g, (_, type, sel) => {
					let ret = ` ${type}="`;

					if (type == 'class') {
						sel.split(/\s+/).forEach(klass => {
							klass = klass.trim();
							if (klass) ret += selectors.minifySelector(klass, true) + ' ';
						});
					} else {
						sel = sel.trim();
						if (sel) ret += selectors.minifySelector(sel, false);
					}

					ret += '"';
					return ret;
				});
			} else if (~FILE_EXTENSIONS.xml.indexOf(extension)) {
			} else {
				throw new Error(`Unknown file type for ${file}`);
			}
		}

		codeToMinify.push({
			type: extension,
			path: file,
			code: data
		});
	});

	if (MINIFY_SELECTORS && settings.exportSelectors) {
		let json = JSON.stringify(selectors.export()),
			file = `./${settings.exportSelectors}.json`;
		fs.writeFileSync(file, json);
		console.log('Exported selectors to %s', file);
	}

	// Then, minify all the files and write them to their destinations
	codeToMinify.forEach(({ type, path, code }) => {
		let folder = path.split('/').slice(0, -1).join('/');

		if (~FILE_EXTENSIONS.js.indexOf(type)) {
			if (MINIFY_JS) code = minifyJS(code, { debug: DEBUG });
		} else if (~FILE_EXTENSIONS.css.indexOf(type)) {
			code = minifyCSS(code);
		} else if (~FILE_EXTENSIONS.html.indexOf(type)) {
			code = minifyHTML(code, {
				minifyJS: MINIFY_HTML_JS,
				minifyCSS: MINIFY_HTML_CSS,
			});
		} else if (~FILE_EXTENSIONS.xml.indexOf(type)) {
			code = minifyXML(code);
		}

		if (settings.onminify) code = settings.onminify(code, type, path) || code;

		mkdirp.sync(DST_DIR + folder);
		fs.writeFileSync(DST_DIR + path, code);
		console.log(`Processed ${path}`);
	});
};
